// Copyright (C) 2013 by - FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
//                       - Institut Eurécom
//                       - Télécom ParisTech
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoARDrone.
// 
// demoARDrone is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// demoARDrone is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with demoARDrone. If not, see 
// <http://www.gnu.org/licenses/>.


// fly like an insect
// ==================

#include "appInsectoid.h"
#include "hawaii/common/helpers.h"
#include "matcher.h"

// custom initialization of wrapped classes
DroneAppInsectoid::DroneAppInsectoid() :
	headingTarget(  0.5 ),
	altitudeTarget( 0.5 ) {
	Matcher::parameters params ;
	params.nms_n                  =   3 ; // default   3, to be tuned
	params.nms_tau                =  50 ; // default  50, to be tuned
	params.match_binsize          =  50 ; // default  50, to be tuned
	params.match_radius           =  96 ; // default 200, to be tuned
	params.outlier_flow_tolerance =   5 ; // default   3, to be tuned
	params.multi_stage            =   1 ; // default   1, to be tuned
	params.half_resolution        =   1 ; // default   1, to be tuned
	params.refinement             =   0 ; // default   1
	this->matcherPtr.reset( new Matcher( params ) ) ;
}

// generate targets from forward camera images
void DroneAppInsectoid::processImageFront( const cv_bridge::CvImage imageFront ) {
	
	// convert input to grayscale, keep a copy for visualization
	cv::Mat imageGray ;
	cv::cvtColor( imageFront.image, imageGray, cv::COLOR_BGR2GRAY ) ;
	cv::Mat visu ;
	imageFront.image.copyTo( visu ) ;
	
	// find optical flow matches
	int32_t dims[] = { imageGray.cols,
	                   imageGray.rows,
	                   imageGray.step } ;
	this->matcherPtr->pushBack( imageGray.data, dims, false ) ;
	this->matcherPtr->matchFeatures( 0 ) ; // "0" as in "optical flow, but no stereo"
	const std::vector< Matcher::p_match > matches = this->matcherPtr->getMatches() ;
	
	// draw the matches
	for( const auto& match : matches ) {
		cv::line( visu,
		          cv::Point( match.u1c, match.v1c ),
		          cv::Point( match.u1p, match.v1p ),
		          cv::Scalar( 0, 127, 255 ), 1, 4 ) ;
	}
	
	// sort flow vector lengths into a grid
	static const size_t gridRows = 3,
	                    gridCols = 5 ;
	std::vector< float > flowsAll[ gridRows ][ gridCols ] ;
	for( const auto& match : matches ) {
		
		// compute length
		const float du = match.u1c - match.u1p,
		            dv = match.v1c - match.v1p ;
		const float length = sqrtf( du*du + dv*dv ) ;
		
		// add to corresponding grid cell
		const size_t gridRow = floorf( match.v1c / imageGray.rows * gridRows ),
		             gridCol = floorf( match.u1c / imageGray.cols * gridCols ) ;
		flowsAll[ gridRow ][ gridCol ].push_back( length ) ;
	}
	
	// find representative (75th percentile) length for each grid cell, keep some statistics
	float flows[ gridRows ][ gridCols ] ;
	float flowMin =   INFINITY,
	      flowMax = - INFINITY ;
	size_t flowsEmpty = 0 ;
	for( size_t row = 0 ; row < gridRows ; ++row ) {
	for( size_t col = 0 ; col < gridCols ; ++col ) {
		std::vector< float >& bin = flowsAll[ row ][ col ] ;
		if( bin.empty() ) {
			flows[ row ][ col ] = NAN ;
			++flowsEmpty ;
			continue ;
		}
		std::sort( bin.begin(), bin.end() ) ;
		const float flowCurr = flows[ row ][ col ] = bin[ bin.size() * 3 / 4 ] ;
		if( flowMin > flowCurr ) { flowMin = flowCurr ; }
		if( flowMax < flowCurr ) { flowMax = flowCurr ; }
	} }
	if( flowsEmpty > 12 ) { return ; }
	
	// normalize all bins' representative length
	for( size_t row = 0 ; row < gridRows ; ++row ) {
	for( size_t col = 0 ; col < gridCols ; ++col ) {
		float& flow = flows[ row ][ col ] ;
		flow = ( isnanf( flow ) ? 0.0f : ( flow - flowMin ) / ( flowMax - flowMin ) ) ;
	} }
	
	// use normalized lengths as weights to determine target flight direction
	double uTarget = 0.0f,
	       vTarget = 0.0f ;
	double weightSum = 0.0f ;
	for( size_t row = 0 ; row < gridRows ; ++row ) {
	for( size_t col = 0 ; col < gridCols ; ++col ) {
		const double weight = pow( 1.0f - flows[ row ][ col ], 3.0 ) ;
		weightSum += weight ;
		uTarget += col * weight ;
		vTarget += row * weight ;
	} }
	uTarget /= weightSum ; uTarget += 0.5 ; uTarget *= (double)imageGray.cols / (double)gridCols ;
	vTarget /= weightSum ; vTarget += 0.5 ; vTarget *= (double)imageGray.rows / (double)gridRows ;
	
	// draw a circle at the target direction
	cv::circle( visu,
	            cv::Point( uTarget, vTarget ),
	            20,
	            cv::Scalar( 255, 127, 0 ),
	            4                             ) ;
	
	// convert u-coordinate to target heading
	const double headingError = atan2( this->principalPointFrontU - uTarget, this->focalLengthFrontU ) ;
	double headingTargetTemp = this->navdata.rotZ / 180.0 * CV_PI + headingError ;
	while( headingTargetTemp <  - CV_PI ) { headingTargetTemp += 2.0 * CV_PI ; }
	while( headingTargetTemp >=   CV_PI ) { headingTargetTemp -= 2.0 * CV_PI ; }
	if( !isnan( headingTargetTemp ) ) { this->headingTarget( headingTargetTemp ) ; }
	
	// TODO convert v-coordinate to target altitude
	this->altitudeTarget( 1.0 ) ;
	
	// show the visualization
	HAWAII_IMSHOW( visu ) ;
	cv::waitKey( 1 ) ;
}

// perform higher-rate closed loop control with on-board measurements
void DroneAppInsectoid::processNavdata( const ardrone_autonomy::Navdata navdata ) {
	
	// heading control
	if( isnan( this->headingTarget ) ) {
		this->commands.movement.angular.z = 0.0 ;
	} else {
		double headingError = this->headingTarget - navdata.rotZ / 180.0 * CV_PI ;
		while( headingError <  - CV_PI ) { headingError += 2.0 * CV_PI ; }
		while( headingError >=   CV_PI ) { headingError -= 2.0 * CV_PI ; }
		this->commands.controlAngZ( headingError ) ;
	}
	
	// altitude control
	if( isnan( this->altitudeTarget ) ) {
		this->commands.movement.linear.z = 0.0 ;
	} else {
		this->commands.controlLinZ( this->altitudeTarget - navdata.altd / 1000.0 ) ;
	}
	
	// common movement potential for forward and sideways motion, sideways speed coupled to yaw rate
	constexpr double moveXY = 0.2 ;
	const double moveY = this->commands.movement.linear.y = moveXY * this->commands.movement.angular.z ;
	                     this->commands.movement.linear.x = sqrt( moveXY*moveXY - moveY*moveY ) ;
}
