// Copyright (C) 2013 by - FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
//                       - Institut Eurécom
//                       - Télécom ParisTech
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoARDrone.
// 
// demoARDrone is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// demoARDrone is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with demoARDrone. If not, see 
// <http://www.gnu.org/licenses/>.


// monocular dense 3D reconstruction (from images before and after height change)
// ==============================================================================

#include "dense3D.h"
#include "odometryDrone.h"
#include "hawaii/common/helpers.h"
#include "hawaii/common/error.h"
#include "viso_mono.h"
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>

// helpers only used here
namespace {
	
	// align images captured while hovering in place, use differences as cues for moving objects
	void detectMotion( const cv::Mat  imageA,
	                   const cv::Mat  imageB,
	                         cv::Mat& motionMaskA ) {
		// check input
		HAWAII_ERROR_CONDITIONAL( imageA.empty()
		                       || imageB.empty(),
		                          "Input images must not be empty." ) ;
		HAWAII_ERROR_CONDITIONAL( imageA.type() != CV_8UC1
		                       || imageB.type() != CV_8UC1,
		                          "Input image type must be \"CV_8UC1\"." ) ;
		HAWAII_ERROR_CONDITIONAL( imageA.size() != imageB.size(),
		                          "input image sizes must be equal." ) ;
		
		// find "libviso2" matches
		Matcher::parameters params ;
		params.match_binsize   =  32 ; //  50 TODO test
		params.match_radius    = 100 ; // 200
		params.half_resolution =   1 ; //   1
		params.refinement      =   2 ; //   1
		Matcher matcher( params ) ;
		int32_t dimsA[] = { imageA.cols, imageA.rows, imageA.step } ;
		int32_t dimsB[] = { imageB.cols, imageB.rows, imageB.step } ;
		matcher.pushBack( imageA.data, dimsA, false ) ;
		matcher.pushBack( imageB.data, dimsB, false ) ;
		matcher.matchFeatures( 0 ) ;
		std::vector< Matcher::p_match > matches = matcher.getMatches() ;
		
		// visualize the matches
/*		cv::Mat blend ; cv::addWeighted( imageA, 0.5, imageB, 0.5, 0.0, blend ) ;
		cv::Mat visuAlign ; cv::cvtColor( blend, visuAlign, CV_GRAY2BGR ) ;
		for( const auto& match : matches ) {
			cv::line( visuAlign,
			          cv::Point2f( match.u1p, match.v1p ),
			          cv::Point2f( match.u1c, match.v1c ),
			          cv::Scalar( 255, 0, 255 ), 1        ) ;
		}
		HAWAII_IMSHOW( visuAlign ) ; //*/
		
		// convert the matches to "OpenCV" points
		std::vector< cv::Point2f > pointsA,
		                           pointsB ;
		for( const auto& match : matches ) {
			pointsA.push_back( cv::Point2f( match.u1p, match.v1p ) ) ;
			pointsB.push_back( cv::Point2f( match.u1c, match.v1c ) ) ;
		}
		
		// find and apply homography via "OpenCV"
		cv::Mat homography = cv::findHomography( pointsB, // source plane
		                                         pointsA, // target plane
															  cv::RANSAC, 3.0 ) ;
		cv::Mat stabilizedB ;
		cv::warpPerspective( imageB, stabilizedB,
		                     homography,
		                     imageB.size(),
		                     cv::INTER_LINEAR,
		                     cv::BORDER_REPLICATE ) ;
//		HAWAII_IMSHOW( stabilizedB ) ;
		
		// simply use the absolute difference as a mask for moving objects
		cv::absdiff( imageA, stabilizedB, motionMaskA ) ;
		
	} // function "detectMotion()"
	
	// rectify images captured before and after the height change
	void rectify( const cv::Matx33d camera,
	              const cv::Vec3d   rotationDrone,
	              const cv::Vec3d   translationDrone,
	              const cv::Mat     imageBefore,
	              const cv::Mat     imageAfter,
	                    cv::Mat&    rectifiedBefore,
	                    cv::Mat&    rectifiedAfter,
	                    cv::Mat&    homographyAfter  ) {
		
		// instantiate visual odometry with camera and ground plane parameters
		VisualOdometryMono::parameters params ;
		params.match.match_binsize   =  32 ; //  50 TODO test
		params.match.match_radius    = 200 ; // 200
		params.match.half_resolution =   1 ; //   1
		params.match.refinement      =   2 ; //   1
		params.bucket.max_features  =  4 ; //  2 TODO test
		params.bucket.bucket_width  = 32 ; // 50 TODO test
		params.bucket.bucket_height = 32 ; // 50 TODO test
		params.calib.f  = camera( 0, 0 ) ;
		params.calib.cu = camera( 0, 2 ) ;
		params.calib.cv = camera( 1, 2 ) ;
		params.height = 0.04 - translationDrone( 1 ) ; // sign different from default "computer vision coordinates", camera higher than sensor
      params.pitch  = rotationDrone( 0 ) ;
      params.roll   = rotationDrone( 2 ) ;
		params.inlier_threshold =   0.00010 ; //   0.00001
		params.motion_threshold = 150.0     ; // 100.0
		VisualOdometryMono viso( params ) ;
		
		// estimate relative position between the images
		int32_t dimsBefore[] = { imageBefore.cols, imageBefore.rows, imageBefore.step } ;
		int32_t dimsAfter[]  = { imageAfter.cols,  imageAfter.rows,  imageAfter.step  } ;
		viso.process( imageBefore.data, dimsBefore, false ) ;
		viso.process( imageAfter.data,  dimsAfter,  false ) ;
		const Matrix motionViso = viso.getMotion() ;
		
		// visualize the matches
/*		cv::Mat blend ; cv::addWeighted( imageBefore, 0.5, imageAfter, 0.5, 0.0, blend ) ;
		cv::Mat visuRectify ; cv::cvtColor( blend, visuRectify, CV_GRAY2BGR ) ;
		for( const auto& match : viso.getMatches() ) {
			cv::line( visuRectify,
			          cv::Point2f( match.u1p, match.v1p ),
			          cv::Point2f( match.u1c, match.v1c ),
			          cv::Scalar( 255, 0, 255 ), 1        ) ;
		}
		HAWAII_IMSHOW( visuRectify ) ; //*/
		
		// convert to "OpenCV" format
		cv::Matx33d rotationViso    ;
		cv::Vec3d   translationViso ;
		rotationViso(0,0)  =   motionViso.val[0][0] ;
		rotationViso(0,1)  =   motionViso.val[0][1] ;
		rotationViso(0,2)  =   motionViso.val[0][2] ;
		rotationViso(1,0)  =   motionViso.val[1][0] ;
		rotationViso(1,1)  =   motionViso.val[1][1] ;
		rotationViso(1,2)  =   motionViso.val[1][2] ;
		rotationViso(2,0)  =   motionViso.val[2][0] ;
		rotationViso(2,1)  =   motionViso.val[2][1] ;
		rotationViso(2,2)  =   motionViso.val[2][2] ;
		translationViso(0) = - motionViso.val[0][3] ;
		translationViso(1) = - motionViso.val[1][3] ;
		translationViso(2) = - motionViso.val[2][3] ;
//		HAWAII_PRINT( rotationViso    ) ;
//		HAWAII_PRINT( translationViso ) ;
		
		// rectify pair of frames
		static const cv::Vec< double, 5 > distortion( 0.0, 0.0, 0.0, 0.0, 0.0 ) ;
		cv::Mat homographyBefore/*,
				  homographyAfter */ ;
		cv::Mat projBefore, projAfter, reproj ;
		cv::stereoRectify( camera, distortion, camera, distortion, imageAfter.size(),
		                   rotationViso, translationViso,
		                   homographyBefore, homographyAfter,
		                   projBefore, projAfter, reproj ) ;
//		HAWAII_PRINT( homographyAfter  ) ;
//		HAWAII_PRINT( homographyBefore ) ;
		
		homographyBefore = projBefore.colRange( 0, 3 ) * homographyBefore * (cv::Mat)camera.inv() ;
		homographyAfter  = projAfter.colRange(  0, 3 ) * homographyAfter  * (cv::Mat)camera.inv() ;
		cv::warpPerspective( imageBefore, rectifiedBefore, homographyBefore, imageBefore.size(), cv::INTER_LINEAR, cv::BORDER_REPLICATE, 0 ) ;
		cv::warpPerspective( imageAfter,  rectifiedAfter,  homographyAfter,  imageAfter.size(),  cv::INTER_LINEAR, cv::BORDER_REPLICATE, 0 ) ;
//		HAWAII_IMSHOW( rectifiedBefore ) ;
//		HAWAII_IMSHOW( rectifiedAfter  ) ;
		
		// "traditional" way, but why create maps for one-time use?
/*		cv::Mat rectifiedBeforeMap,
				  rectifiedAfterMap ;
		cv::Mat mapBeforeX, mapBeforeY, mapAfterX, mapAfterY ;
		cv::initUndistortRectifyMap( camera, distortion, homographyBefore, projBefore, imageBefore.size(), CV_32FC1, mapBeforeX, mapBeforeY ) ;
		cv::initUndistortRectifyMap( camera, distortion, homographyAfter,  projAfter,  imageAfter.size(),  CV_32FC1, mapAfterX,  mapAfterY  ) ;
		cv::remap( imageBefore, rectifiedBeforeMap, mapBeforeX, mapBeforeY, cv::INTER_LINEAR ) ;
		cv::remap( imageAfter,  rectifiedAfterMap,  mapAfterX,  mapAfterY,  cv::INTER_LINEAR ) ;
		HAWAII_IMSHOW( rectifiedBeforeMap ) ;
		HAWAII_IMSHOW( rectifiedAfterMap  ) ; //*/
	}
	
	// convert a distance to a color from red over yellow and green to cyan
	class DistanceToColor {
		
		// c'tor using distance limits
		public:
		DistanceToColor( const double distRedArg,
		                 const double distCyanArg ) :
			distRed(  distRedArg  ),
			distCyan( distCyanArg ) {
		}
		protected:
		double distRed,
		       distCyan ;
		
		// convert a distance to a color
		public:
		cv::Vec3b operator ()( double dist ) {
			double distRel = 3.0 * (       dist     - this->distRed )
			                     / ( this->distCyan - this->distRed ) ;
			     if( distRel < 0.0 ) { return cv::Vec3b(                       0,             0,                     255 ) ; } // all red
			else if( distRel < 1.0 ) { return cv::Vec3b(                       0, distRel * 255,                     255 ) ; } // from red to yellow
			else if( distRel < 2.0 ) { return cv::Vec3b(                       0,           255, ( 2.0 - distRel ) * 255 ) ; } // from yellow to green
			else if( distRel < 3.0 ) { return cv::Vec3b( ( distRel - 2.0 ) * 255,           255,                       0 ) ; } // from green to cyan
			else                     { return cv::Vec3b(                     255,           255,                       0 ) ; } // all cyan
		}
	} ;
	
} // anonymous namespace

// c'tor with camera properties
Dense3D::Dense3D( const double focalLength,
                  const double principalPointU,
                  const double principalPointV ) :
	
	// camera matrix
	camera( focalLength, 0.0,         principalPointU,
	        0.0,         focalLength, principalPointV,
	        0.0,         0.0,         1.0             ),
	
	// height change, stabilization mode
	altitudeDelta( 0.15 ),
	stabilization( stabilizationNone ),
	
	// times to hover in place before and after height change, number of images to buffer
	timeHoverBefore( 0.75 ),
	timeHoverAfter(  0.75 ),
	numberHoverBefore( 3 ),
	numberHoverAfter(  3 ),
	indexReferenceBefore( numberHoverBefore / 2 ),
	indexReferenceAfter(  numberHoverAfter  / 2 ),
	
	// internal state
	state( stateInactive ),
	
	// forward/backward and left/right motions during the height change
	errorVisoForward( 0.05 ),
	errorVisoLeft(    0.05 ),
	errorVisoValid( false ) {
}

// buffer images and poses from drone's on-board odometry before and after height change
void Dense3D::processImageFront( const cv::Mat   image,
                                 const cv::Vec3d rotationGlobal,
                                 const cv::Vec3d translationGlobal ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( image.empty(),
	                          "Image must not be empty." ) ;
	HAWAII_ERROR_CONDITIONAL( image.type() != CV_8UC1
	                       && image.type() != CV_8UC3,
	                          "Image type must be \"CV_8UC1\" or \"CV_8UC3\"." ) ;
	
	// convert to grayscale if necessary
	cv::Mat imageGray ;
	if( image.type() == CV_8UC1 ) { imageGray = image ; }
	else { cv::cvtColor( image, imageGray, cv::COLOR_BGR2GRAY ) ; }
	
	std::cout <<this->state;
	// buffer some images while hovering before a height change
	//if( this->state == stateBeforeHeightChange) {
	//	std::cout << "INFO: dense 3D: problem, can't access state HEIGHT CHANGE" << std::endl ;
	//}
	if( this->state == stateBeforeHeightChange
	 && ( this->timerHover.toc() / 1000.0 > this->timeHoverBefore / this->numberHoverBefore ) ) {
		
		// buffer all data
		this->imagesBefore.push_back(       imageGray         ) ;
		this->rotationsBefore.push_back(    rotationGlobal    ) ;
		this->translationsBefore.push_back( translationGlobal ) ;
		this->timerHover.tic() ;
		std::cout << "INFO: dense 3D: buffered image BEFORE height change" << std::endl ;
		
		// initialize visual odometry with reference image before height change
		if( this->stabilization == stabilizationVisual
		 && this->imagesBefore.size() - 1 == this->indexReferenceBefore ) {
			this->visoInit( imageGray,
			                rotationGlobal,
			                translationGlobal ) ;
		}
			
		// once buffer is full...
		if( this->imagesBefore.size() >= this->numberHoverBefore ) {
			
			// change state
			this->state = this->stateDuringHeightChange ;
			std::cout << "INFO: dense 3D: changed state to HEIGHT CHANGE" << std::endl ;
		}
	}
	
	// while performing height change...
	if( this->state == stateDuringHeightChange ) {
		
		// use visual odometry to update relative pose w.r.t. last image before height change
		if( this->stabilization == stabilizationVisual ) {
			this->visoProcess( imageGray,
			                   rotationGlobal,
			                   translationGlobal ) ;
		}
	}
	
	// buffer some images while hovering after a height change
	if( this->state == stateAfterHeightChange
	 && ( this->timerHover.toc() / 1000.0 > this->timeHoverAfter / this->numberHoverAfter ) ) {
		
		// buffer all data
		this->imagesAfter.push_back(       imageGray             ) ;
		this->rotationsAfter.push_back(    rotationGlobal    ) ;
		this->translationsAfter.push_back( translationGlobal ) ;
		this->timerHover.tic() ;
		std::cout << "INFO: dense 3D: buffered image AFTER height change" << std::endl ;
		
		// once buffer is full...
		if( this->imagesAfter.size() >= this->numberHoverAfter ) {
			
			// change state
			this->state = this->stateProcessing ;
			std::cout << "INFO: dense 3D: changed state to PROCESSING BUFFERS" << std::endl ;
			
			// launch thread to asynchronously process the buffered data
			this->threadComputeResults = boost::thread( &Dense3D::computeResults, this ) ;
		}
	}
	
} // method "Dense3D::processImageFront()"

// get control commands depending on state
bool Dense3D::getCommands(       DroneCommands& commands,
                           const cv::Vec3d      rotationGlobal,
                           const cv::Vec3d      translationGlobal ) {
	
	// handle all states independently
	switch( this->state ) {
			
		// hover in place
		case stateBeforeHeightChange :
		case stateAfterHeightChange  : {
			
			// engage on-board stabilization
			commands.hover() ;
			
		} break ;
		
		// build a vertical stereo camera by changing height
		case stateDuringHeightChange : {
			
			// closed-loop control of forward/backward and left/right motions based on visual odometry
			double errorForward = 0.0,
			       errorLeft    = 0.0 ;
			switch( this->stabilization ) {
				
				// try not to move forward/backward or left/right via open-loop control
				case stabilizationNone : {
					commands.movement.linear.x = 0.0f ;
					commands.movement.linear.y = 0.0f ;
				} break ;
				
				// compute and use errors from on-board odometry
				case stabilizationOnboard : {
					const cv::Vec3d translDeltaGlobal = translationGlobal - this->translationsBefore[ this->indexReferenceBefore ] ;
					const cv::Vec3d translDeltaLocal = OdometryDrone::rotateTranslation( translDeltaGlobal, rotationGlobal( 1 ) ) ;
					errorForward = - translDeltaLocal( 2 ) ;
					errorLeft    =   translDeltaLocal( 0 ) ;
					commands.controlLinX( errorForward ) ;
					commands.controlLinY( errorLeft    ) ;
				} break ;
				
				// use errors from visual odometry
				case stabilizationVisual : {
					errorForward = this->errorVisoForward ;
					errorLeft    = this->errorVisoLeft    ;
					commands.controlLinX( errorForward ) ;
					commands.controlLinY( errorLeft    ) ;
				} break ;
			}
			
			// closed-loop control of altitude and yaw always based on on-board sensors
			const double altitudeBefore = - this->translationsBefore.back()( 1 ),
			             altitudeCurr   = -       translationGlobal(         1 ),
			             yawBefore      = - this->rotationsBefore.back()(    1 ),
			             yawCurr        = -       rotationGlobal(            1 ) ;
			double errorUp  = altitudeBefore - altitudeCurr + this->altitudeDelta,
			       errorYaw = yawBefore      - yawCurr                            ;
			if( errorYaw >   CV_PI ) { errorYaw -= 2.0 * CV_PI ; }
			if( errorYaw < - CV_PI ) { errorYaw += 2.0 * CV_PI ; }
			commands.controlLinZ( errorUp  ) ;
			commands.controlAngZ( errorYaw ) ;
			
			// crank height control up if error is large
			#if 1
				commands.controllerLinZ.disableIntegrator = false ;
				if( 0.0     < this->altitudeDelta
				 && errorUp > this->altitudeDelta * 0.25
				 && commands.movement.linear.z <   0.75f ) {
					commands.controllerLinZ.disableIntegrator = true ;
					commands.movement.linear.z =    0.75f ;
				}
				if( 0.0     > this->altitudeDelta
				 && errorUp < this->altitudeDelta * 0.25
				 && commands.movement.linear.z > - 0.75f ) {
					commands.controllerLinZ.disableIntegrator = true ;
					commands.movement.linear.z =  - 0.75f ;
				}
			#endif
			
			// show the errors
			HAWAII_PRINT( errorForward ) ;
			HAWAII_PRINT( errorLeft    ) ;
			HAWAII_PRINT( errorUp      ) ;
			HAWAII_PRINT( errorYaw     ) ;
			std::cout << std::endl ; //*/
			
			// suppress any specific flight maneuver
			commands.maneuver = DroneCommands::maneuverNone ;
			
			// check if height change has been completed
			if( ( this->stabilization != stabilizationVisual || this->errorVisoValid )
			 && ( this->stabilization == stabilizationNone   || fabs( errorForward ) < 0.20 * this->altitudeDelta ) // greater impact than others
			 && ( this->stabilization == stabilizationNone   || fabs( errorLeft    ) < 0.20 * this->altitudeDelta )
			 &&                                                 fabs( errorUp      ) < 0.20 * this->altitudeDelta
			 &&                                                 fabs( errorYaw     ) < 5.0 / 180.0 * CV_PI ) {
				
				// change state
				this->state = stateAfterHeightChange ;
				this->timerHover.tic() ;
				std::cout << "INFO: dense 3D: changed state to HOVER AFTER height change" << std::endl ;
			}
			
		} break ;
		
		// do not modify "commands"
		case stateProcessing :
		case stateInactive   : {
			
		} break ;
		
	} // switch "all states"
	
	// return "true" only if "commands" has been modified
	return ( this->state != stateInactive
	      && this->state != stateProcessing ) ;
	
} // method "Dense3D::getCommands()"

// trigger the scan flight maneuver
bool Dense3D::trigger() {
	
	// indicate that still busy with previous scan
	if( this->state == stateProcessing ) {
		return false ;
	}
	
	// clear all buffers and results for next iteration
	this->imagesBefore.clear() ;
	this->imagesAfter.clear()  ;
	this->rotationsBefore.clear() ;
	this->rotationsAfter.clear()  ;
	this->translationsBefore.clear() ;
	this->translationsAfter.clear()  ;
	this->distance.release() ;
	this->motionMask.release() ;
	
	// change state
	this->state = stateBeforeHeightChange ;
	this->timerHover.tic() ;
	std::cout << "INFO: dense 3D: changed state to HOVER BEFORE height change" << std::endl ;
	return true ;
}

// return distance map and moving object mask
bool Dense3D::getResults( cv::Mat& distanceArg,
                          cv::Mat& motionMaskArg ) const {
	
	// only return previously computed results while next scan not yet triggered
	if( this->state != stateProcessing
	 && !this->distance.empty()
	 && !this->motionMask.empty() ) {
		
		// set arguments
		distanceArg   = this->distance   ;
		motionMaskArg = this->motionMask ;
		return true ;
	}
	
	// indicate unmodified arguments
	return false ;
}

// compute distance map and moving object mask
void Dense3D::computeResults() {
	
	// detect motion w.r.t last image before and first image after height change
	cv::Mat motionBefore2, motionAfter0 ;
	detectMotion( this->imagesBefore[ 2 ], this->imagesBefore[ 0 ], motionBefore2 ) ;
	detectMotion( this->imagesAfter[  0 ], this->imagesAfter[  2 ], motionAfter0  ) ;
/*	HAWAII_IMSHOW( motionBefore2 ) ;
	HAWAII_IMSHOW( motionAfter0  ) ; //*/
	
	// rectify last image before and first image after height change
	cv::Mat rectifiedBefore2, rectifiedAfter0, homographyAfter ;
	rectify( this->camera,
	         this->rotationsAfter[ 0 ], this->translationsAfter[ 0 ],
	         this->imagesBefore[ 2 ], this->imagesAfter[ 0 ],
	         rectifiedBefore2, rectifiedAfter0,
	         homographyAfter                                         ) ;
	
	// 3D reconstruction
	cv::StereoSGBM sgbm ;
	sgbm.minDisparity        =   0 ;
	sgbm.numberOfDisparities =  64 ;
	sgbm.SADWindowSize = 7 ;
	sgbm.P1 =  8 * sgbm.SADWindowSize * sgbm.SADWindowSize ;
	sgbm.P2 = 32 * sgbm.SADWindowSize * sgbm.SADWindowSize ;
	sgbm.disp12MaxDiff = 2 ; // 1
	sgbm.preFilterCap = 127 ; // 63
	sgbm.uniquenessRatio = 20 ; // 10
	sgbm.speckleWindowSize = 200 ; // 100
	sgbm.speckleRange = 2 ; // 32
	sgbm.fullDP = true ;
	
	double t = sqrt( ( translationsAfter[ 0 ][ 0 ] - translationsBefore[ 2 ][ 0 ] ) * ( translationsAfter[ 0 ][ 0 ] - translationsBefore[ 2 ][ 0 ] )
	               + ( translationsAfter[ 0 ][ 1 ] - translationsBefore[ 2 ][ 1 ] ) * ( translationsAfter[ 0 ][ 1 ] - translationsBefore[ 2 ][ 1 ] )
	               + ( translationsAfter[ 0 ][ 2 ] - translationsBefore[ 2 ][ 2 ] ) * ( translationsAfter[ 0 ][ 2 ] - translationsBefore[ 2 ][ 2 ] ) ) ;
	HAWAII_PRINT( t ) ;
	
	cv::Mat disp ; sgbm( rectifiedAfter0.t(), rectifiedBefore2.t(), disp ) ;
	cv::Mat visu( disp.size(), CV_8UC3 ) ;
	DistanceToColor distanceToColor(  1.0, 20.0 ) ;
	for( int row = 0 ; row < visu.rows ; ++row ) {
	for( int col = 0 ; col < visu.cols ; ++col ) {
		if( disp.at< int16_t >( row, col ) <= 0 ) {
			visu.at< cv::Vec3b >( row, col ) = cv::Vec3b::all( 0 ) ;
		} else {
			visu.at< cv::Vec3b >( row, col ) = distanceToColor( t * 565.0 / (double)( disp.at< int16_t >( row, col ) / ( 16 - 4 ) ) ) ;
		}
	} }
	cv::Mat blend2 ; cv::addWeighted( rectifiedBefore2, 0.5, rectifiedAfter0, 0.5, 0.0, blend2 ) ;
	HAWAII_IMSHOW( blend2   ) ; cv::imwrite( "/home/benjamin/Desktop/dense/Blend.png", blend2   ) ;
	HAWAII_IMSHOW( visu.t() ) ; cv::imwrite( "/home/benjamin/Desktop/dense/Visu.png",  visu.t() ) ;
	cv::warpPerspective( visu.t(), visu,
	                     homographyAfter,
	                     visu.t().size(),
	                     cv::INTER_LINEAR | cv::WARP_INVERSE_MAP,
	                     cv::BORDER_REPLICATE ) ;
	HAWAII_IMSHOW( visu ) ;
	
	// force screen to be updated
	cv::waitKey( 1 ) ;
	
	// change state when done
	this->state = stateInactive ;
	std::cout << "INFO: dense 3D: changed state to INACTIVE" << std::endl ;
}

// initialize visual odometry with last image before height change
void Dense3D::visoInit( const cv::Mat   imageBefore,
                        const cv::Vec3d rotationGlobal,
                        const cv::Vec3d translationGlobal ) {
	
	// instantiate wrapped visual odometry with required parameters
	// TODO tune remaining parameters, also apply to "sparse3D.cpp"
	VisualOdometryMono::parameters params ;
	params.ransac_iters     = 1000       ; // default 2000, to be tuned
	params.inlier_threshold =    0.00010 ; // default    0.00001
	params.motion_threshold =  100.0     ; // default  100.0
	params.match.nms_n                  =   3 ; // default   3, to be tuned
	params.match.nms_tau                =  25 ; // default  50, to be tuned
	params.match.match_binsize          =  32 ; // default  50, to be tuned
	params.match.match_radius           = 384 ; // default 200, to be tuned
	params.match.outlier_flow_tolerance =   3 ; // default   3, to be tuned
	params.match.multi_stage            =   1 ; // default   1, to be tuned
	params.match.half_resolution        =   1 ; // default   1, to be tuned
	params.match.refinement             =   2 ; // default   1
	params.bucket.max_features  =  2 ; // default  2, to be tuned
	params.bucket.bucket_width  = 32 ; // default 50, to be tuned
	params.bucket.bucket_height = 32 ; // default 50, to be tuned
	params.calib.f  = 0.5 * ( this->camera( 0, 0 )
	                        + this->camera( 1, 1 ) ) ; // (to be calibrated, not tuned!)
	params.calib.cu = this->camera( 0, 2 ) ;           // (to be calibrated, not tuned!)
	params.calib.cv = this->camera( 1, 2 ) ;           // (to be calibrated, not tuned!)
	this->visoPtr.reset( new VisualOdometryMono( params ) ) ;
	
	// update ground plane parameters used to resolve the monocular scale ambiguity
	this->visoPtr->param.height = 0.04 - translationGlobal( 1 ) ; // sign different from default "computer vision coordinates", camera higher than sensor
	this->visoPtr->param.pitch = rotationGlobal( 0 ) ;
	this->visoPtr->param.roll  = rotationGlobal( 2 ) ;
	
	// add the image as both previous and current for good measure
	int32_t dims[] = { imageBefore.cols,
	                   imageBefore.rows,
	                   imageBefore.step } ;
	this->visoPtr->process( imageBefore.data, dims, false ) ;
	this->visoPtr->process( imageBefore.data, dims, false ) ;
}

// compute relative position between last image before and each image during height change
void Dense3D::visoProcess( const cv::Mat   imageDuring,
                           const cv::Vec3d rotationGlobal,
                           const cv::Vec3d translationGlobal ) {
	
	// update ground plane parameters used to resolve the monocular scale ambiguity
	this->visoPtr->param.height = 0.04 - translationGlobal( 1 ) ; // sign different from default "computer vision coordinates", camera higher than sensor
	this->visoPtr->param.pitch = rotationGlobal( 0 ) ;
	this->visoPtr->param.roll  = rotationGlobal( 2 ) ;
	
	// always keep the initially set last image before the height change
	int32_t dims[] = { imageDuring.cols,
	                   imageDuring.rows,
	                   imageDuring.step } ;
	if( this->visoPtr->process( imageDuring.data, dims, true ) ) {
		
		// if odometry has been successful, get the relative motion w.r.t. the last image before the height change
		const Matrix motionDeltaViso = this->visoPtr->getMotion() ;
		
		// use on-board odometry to correct the scale: While on-board odometry is more susceptible to noise, visual 
		//                                             odometry may pick a wrong set of 3D points when fitting the 
		//                                             ground plane. Therefore, we fully trust visual odometry if its 
		//                                             translation is similar enough to on-board odometry. Otherwise we 
		//                                             scale visual odometry results to make translation lengths or 
		//                                             altitude components match.
		// developer note: Almost the same code exists in "Sparse3D::processImageFront()" - keep them synchronized!
		const cv::Vec3d translDeltaGlobalOnboard = translationGlobal - this->translationsBefore.back() ;
		const double lengthOnboard = sqrt( translDeltaGlobalOnboard( 0 ) * translDeltaGlobalOnboard( 0 )
		                                 + translDeltaGlobalOnboard( 1 ) * translDeltaGlobalOnboard( 1 )
		                                 + translDeltaGlobalOnboard( 2 ) * translDeltaGlobalOnboard( 2 ) ),
		             lengthViso    = sqrt( motionDeltaViso.val[ 0 ][ 3 ] * motionDeltaViso.val[ 0 ][ 3 ]
		                                 + motionDeltaViso.val[ 1 ][ 3 ] * motionDeltaViso.val[ 1 ][ 3 ]
		                                 + motionDeltaViso.val[ 2 ][ 3 ] * motionDeltaViso.val[ 2 ][ 3 ] ) ;
		double scaleFactor = lengthOnboard / lengthViso ;
		if( scaleFactor > 0.9
		 && scaleFactor < 1.1 ) {
			scaleFactor = 1.0 ;
		}
		this->errorVisoForward(   motionDeltaViso.val[ 2 ][ 3 ] * scaleFactor ) ;
		this->errorVisoLeft(    - motionDeltaViso.val[ 0 ][ 3 ] * scaleFactor ) ;
		this->errorVisoValid = true ;
		std::cout << "INFO: dense 3D: visual odometry SUCCEEDED"
//		          << ", scaleFactor = " << scaleFactor
//		          << ", lengthOnboard = " << lengthOnboard
//		          << ", lengthViso = " << lengthViso
		          << std::endl ;
		
		// visualize results
/*		cv::Mat visu ;
		cv::addWeighted( imageDuring, 0.5, this->imagesBefore.back(), 0.5, 0.0, visu ) ;
		cv::cvtColor( visu, visu, CV_GRAY2BGR ) ;
		std::vector< Matcher::p_match > matches = this->visoPtr->getMatches() ;
		for( const auto& match : matches ) {
			cv::line( visu,
			          cv::Point( match.u1c, match.v1c ),
			          cv::Point( match.u1p, match.v1p ),
			          cv::Scalar( 0, 127, 255 ), 2, 4 ) ;
		}
		HAWAII_IMSHOW( visu ) ; //*/
		
	} else {
		
		// indicate if odometry has failed, fade the error trackers down towards zero
		this->errorVisoValid = false ;
		this->errorVisoForward( 0.0 ) ;
		this->errorVisoLeft(    0.0 ) ;
		std::cout << "INFO: dense 3D: visual odometry FAILED" << std::endl ;
	}
	
} // method "Dense3D::visoProcess()"
