/*
 * message_data.h
 *
 *  Created on: Aug 30, 2013
 *      Author: truongnt
 */

#ifndef MESSAGE_DATA_H_
#define MESSAGE_DATA_H_

#include <opencv2/opencv.hpp>
#include <opencv2/core/core.hpp>
#include <iostream>

//https://groups.google.com/forum/#!topic/boost-list/15nKgxGFAeQ
#include <boost/date_time/posix_time/time_serialize.hpp>

using namespace cv;
//-----------------------------------------------------------------------------------------------------------------------------
BOOST_SERIALIZATION_SPLIT_FREE(cv::Vec3d)
namespace boost {
	namespace serialization {

	/*** Vec3d ***/
	template<class Archive>
	//the reason version parameter exist to support the compatibility of system,
	//in some case, we can still working with older version of save and load.
	//default value for version is 0
	//http://www.boost.org/doc/libs/1_37_0/libs/serialization/doc/tutorial.html#versioning
	void save(Archive & ar, const cv::Vec3d& m, const unsigned int version)
	{
		if (version == 0) {
			ar & m.val[0];
			ar & m.val[1];
			ar & m.val[2];
		}
	}

	template<class Archive>
	void load(Archive & ar, cv::Vec3d& m, const unsigned int version)
	{
		if (version == 0) {
			ar & m.val[0];
			ar & m.val[1];
			ar & m.val[2];
		}
	}

	}
}

//-----------------------------------------------------------------------------------------------------------------------------
enum Command {
	NoCommand,
	ClientSendInfoToServer,
	CloseConnection,
	//remote Drone

	Escape,

	TakeOff,
	Landing,

	Engage,
	DisEngage,

	MoveForward,
	MoveBackward,

	MoveUp,
	MoveDown,

	MoveLeft,
	MoveRight,

	RotateLeft,
	RotateRight,

	//for engaged mode
	AutoMoveForward,
	AutoMoveBackward,

	AutoMoveUp,
	AutoMoveDown,

	AutoMoveLeft,
	AutoMoveRight,

	AutoRotate,
};
class MessageData {
	//only need when we want to access private variables
	//friend class boost::serialization::access;
public:
	cv::Mat mImg;
	Command mCommand;
	int mLapNo;
	boost::posix_time::ptime mTimeStamp;

	cv::Vec3d mRotationVector, mTranslationVector;

	int mCommandIndex;

	MessageData(const MessageData &cSource)
	{
		mImg = cSource.mImg.clone();
		mCommand = cSource.mCommand;
		mLapNo = cSource.mLapNo;
		mTimeStamp = cSource.mTimeStamp;
		mCommandIndex = cSource.mCommandIndex;

		mRotationVector = cSource.mRotationVector;
		mTranslationVector = cSource.mTranslationVector;
	}

	MessageData() {
		mCommandIndex = -1;
		mCommand = Command::NoCommand;
	}

	template <class Archive>
	void serialize(
			Archive& ar,
			unsigned int version
	)
	{
		//std::cout<<"serialization MessageData version="<<version<<std::endl;
		//printing("serialization MessageData version=" + utilities::NumberToString(version));
		if (version == 0) {
			ar & mImg;
			ar & mLapNo;
			ar & mCommand;
			ar & mTimeStamp;

			ar & mCommandIndex;

			ar & mRotationVector(0);
			ar & mRotationVector(1);
			ar & mRotationVector(2);
			ar & mTranslationVector(0);
			ar & mTranslationVector(1);
			ar & mTranslationVector(2);
		}
	}

	bool isCommandMessage(const Command& cm) {
		if (this->mCommand == cm) {
			return true;
		}
		return false;
	}

	static MessageData createMessage(const Command& cm) {
		MessageData msg2;
		msg2.mCommand = cm;
		msg2.mTimeStamp = boost::posix_time::microsec_clock::local_time();
		return msg2;
	}

	static MessageData createMessageCommand(int index, const Command& cm) {
			MessageData msg2;
			msg2.mCommand = cm;
			msg2.mTimeStamp = boost::posix_time::microsec_clock::local_time();
			msg2.mCommandIndex = index;
			return msg2;
		}

public:
//	explicit MessageData() : mImg(), mCommand(), mLapNo(0), mTimeStamp(),
//			mRotationVector(0, 0, 0), mTranslationVector(0, 0, 0),
//			mCommandIndex(0)// explicit to avoid implicit conversion
//	    {}

    void write(FileStorage& fs) const                        //Write serialization for this class
    {
        fs << "{" << "mImg" << this->mImg
        		<< "mLapNo" << this->mLapNo
        		<< "mRotationVector" << this->mRotationVector
        		<< "mTranslationVector" << this->mTranslationVector
        		<< "}";
    }
    void read(const FileNode& node)                          //Read serialization for this class
    {
    	node["mImg"] >> this->mImg;
    	this->mLapNo = (int)node["mLapNo"];
    	this->mRotationVector = (cv::Vec3d)node["mRotationVector"];
    	this->mTranslationVector = (cv::Vec3d)node["mTranslationVector"];
    }
};


//These write and read functions must be defined for the serialization in FileStorage to work
static void write(FileStorage& fs, const std::string&, const MessageData& x)
{
    x.write(fs);
}
static void read(const FileNode& node, MessageData& x, const MessageData& default_value = MessageData()){
    if(node.empty())
        x = default_value;
    else
        x.read(node);
}

#endif /* MESSAGE_DATA_H_ */
